#include <stdio.h>
#include <string.h>

#define TEST_AND_PRINT_EXP(_prefix, _g, _s, _c, _v, _m, _type, _exp) \
	print_result(#_prefix, #_g, #_s, #_c, #_v, #_m, _type, #_exp, __builtin_constant_p(_exp))

#define TEST_AND_PRINT(_prefix, _g, _s, _c, _v, _m, _type) \
	TEST_AND_PRINT_EXP(_prefix, _g, _s, _c, _v, _m, _type, _prefix(_g ## _s ## _c ## _v _m))

#define TEST_AND_PRINT_SET(_type, _prefix, _v, _m)	\
do {							\
	TEST_AND_PRINT(_prefix, g,  ,  , _v, _m, _type);\
	TEST_AND_PRINT(_prefix, g, s,  , _v, _m, _type);\
	TEST_AND_PRINT(_prefix, g,  , c, _v, _m, _type);\
	TEST_AND_PRINT(_prefix, g, s, c, _v, _m, _type);\
	TEST_AND_PRINT(_prefix, l,  ,  , _v, _m, _type);\
	TEST_AND_PRINT(_prefix, l, s,  , _v, _m, _type);\
	TEST_AND_PRINT(_prefix, l,  , c, _v, _m, _type);\
	TEST_AND_PRINT(_prefix, l, s, c, _v, _m, _type);\
} while(0)

#define DECLARE_VARS(_type, _g, _name, _init)		\
	_type _g ## _name = _init;			\
	const _type _g ## c ## _name = _init;		\
	static _type _g ## s ## _name = _init;		\
	static const _type _g ## sc ## _name = _init;

static const char *compiler_string = NULL;

const char* get_padding(unsigned len, unsigned col) {
	static const char *spaces =
		"                                "
		"                                ";
	static const unsigned spaces_size = 64;
	unsigned size = len > col ? 0 : col - len;

	return &spaces[spaces_size - size];
}

void print_result(const char *prefix,
		  const char *g,
		  const char *s,
		  const char *c,
		  const char *v,
		  const char *m,
		  const char *type,
		  const char *exp,
		  int result) {
	if (*prefix) /* no pointers are testing positive, so omitted from output */
		return;
	printf("%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%d\n",
	       compiler_string, prefix,
	       *g == 'g' ? "global" : "local",
	       *s == 's' ? "static" : "",
	       *c == 'c' ? "const"  : "",
	       *v == 's' ? "struct" : v,
	       m, type, exp, result);
}

struct my_struct {
	char c;
	unsigned char uc;
	short s;
	unsigned short us;
	int i;
	unsigned int ui;
	long l;
	unsigned long ul;
	long long ll;
	unsigned long long ull;
	float f;
	double d;
	char *nstr;
	const char *cnstr;
	char *str;
	const char *cstr;
	void *nptr;
	const void *cnptr;
	void *ptr;
	const void *cptr;
};

#define MY_STRUCT_INIT {	\
	.c    = 1,		\
	.uc   = 1u,		\
	.s    = 1,		\
	.us   = 1u,		\
	.i    = 1,		\
	.ui   = 1u,		\
	.l    = 1l,		\
	.ul   = 1ul,		\
	.ll   = 1ll,		\
	.ull  = 1ull,		\
	.f    = 1.0f,		\
	.d    = 1.0,		\
	.nstr  = 0,		\
	.cnstr = 0,		\
	.str   = "str",		\
	.cstr  = "cstr",	\
	.nptr  = 0,		\
	.cnptr = 0,		\
	.ptr   = (void *)4,	\
	.cptr  = (const void *)4\
};


struct my_struct		gs =	MY_STRUCT_INIT
const struct my_struct		gcs =	MY_STRUCT_INIT
static struct my_struct		gss =	MY_STRUCT_INIT
static const struct my_struct	gscs =	MY_STRUCT_INIT

DECLARE_VARS(char,		g, c,     1   )
DECLARE_VARS(unsigned char,	g, uc,    1u  )
DECLARE_VARS(short,		g, S,     1   )
DECLARE_VARS(unsigned short,	g, us,    1u  )
DECLARE_VARS(int,		g, i,     1   )
DECLARE_VARS(unsigned int,	g, ui,    1u  )
DECLARE_VARS(long,		g, l,     1l  )
DECLARE_VARS(unsigned long,	g, ul,    1lu )
DECLARE_VARS(long long,		g, ll,    1ll )
DECLARE_VARS(unsigned long long,g, ull,   1llu)
DECLARE_VARS(float,		g, f,     1.0f)
DECLARE_VARS(double,		g, d,     1.0 )
DECLARE_VARS(char *,		g, nstr,  0       )
DECLARE_VARS(char *,		g, str,   "str"   )
DECLARE_VARS(void *,		g, nptr,  0       )
DECLARE_VARS(void *,		g, ptr,   (void*)4)


int main(int argc, char *argv[]) {
	struct my_struct 		ls =	MY_STRUCT_INIT
	static struct my_struct 	lss =	MY_STRUCT_INIT
	const struct my_struct		lcs =	MY_STRUCT_INIT
	static const struct my_struct 	lscs =	MY_STRUCT_INIT

	DECLARE_VARS(char,		l, c,     1   )
	DECLARE_VARS(unsigned char,	l, uc,    1u  )
	DECLARE_VARS(short,		l, S,     1   )
	DECLARE_VARS(unsigned short,	l, us,    1u  )
	DECLARE_VARS(int,		l, i,     1   )
	DECLARE_VARS(unsigned int,	l, ui,    1u  )
	DECLARE_VARS(long,		l, l,     1l  )
	DECLARE_VARS(unsigned long,	l, ul,    1lu )
	DECLARE_VARS(long long,		l, ll,    1ll )
	DECLARE_VARS(unsigned long long,l, ull,   1llu)
	DECLARE_VARS(float,		l, f,     1.0f)
	DECLARE_VARS(double,		l, d,     1.0 )
	DECLARE_VARS(char *,		l, nstr,  0       )
	DECLARE_VARS(char *,		l, str,   "str"   )
	DECLARE_VARS(void *,		l, nptr,  0       )
	DECLARE_VARS(void *,		l, ptr,   (void*)4)

	if (argc != 2) {
		fprintf(stderr, "Usage: %s <compiler_string>\n", argv[0]);
		return 1;
	}
	compiler_string = argv[1];

	TEST_AND_PRINT_SET("char",		 , s, .c   );
	TEST_AND_PRINT_SET("char*",		&, s, .c   );
	TEST_AND_PRINT_SET("unsigned char",	 , s, .uc  );
	TEST_AND_PRINT_SET("unsigned char*", 	&, s, .uc  );
	TEST_AND_PRINT_SET("short",		 , s, .s   );
	TEST_AND_PRINT_SET("short*",		&, s, .s   );
	TEST_AND_PRINT_SET("short unsigned",	 , s, .us  );
	TEST_AND_PRINT_SET("short unsigned*",	&, s, .us  );
	TEST_AND_PRINT_SET("int",		 , s, .i   );
	TEST_AND_PRINT_SET("int*",		&, s, .i   );
	TEST_AND_PRINT_SET("int unsigned",	 , s, .ui  );
	TEST_AND_PRINT_SET("int unsigned*",	&, s, .ui  );
	TEST_AND_PRINT_SET("long",		 , s, .l   );
	TEST_AND_PRINT_SET("long*",		&, s, .l   );
	TEST_AND_PRINT_SET("long unsigned",	 , s, .ul  );
	TEST_AND_PRINT_SET("long unsigned*",	&, s, .ul  );
	TEST_AND_PRINT_SET("long long",		 , s, .ll  );
	TEST_AND_PRINT_SET("long long*",	&, s, .ll  );
	TEST_AND_PRINT_SET("long long unsigned", , s, .ull );
	TEST_AND_PRINT_SET("long long unsigned*",&,s, .ull );
	TEST_AND_PRINT_SET("float",		 , s, .f   );
	TEST_AND_PRINT_SET("float*",		&, s, .f   );
	TEST_AND_PRINT_SET("double",		 , s, .d   );
	TEST_AND_PRINT_SET("double*",		&, s, .d   );

	TEST_AND_PRINT_SET("char* defre",	 	, s, .str[0]);
	TEST_AND_PRINT_SET("const char* defre",	 	, s, .cstr[0]);
	TEST_AND_PRINT_SET("char* ptr (null)",		, s, .nstr);
	TEST_AND_PRINT_SET("const char* ptr (null)",	, s, .cnstr);
	TEST_AND_PRINT_SET("char* ptr (non-null)",	, s, .str);
	TEST_AND_PRINT_SET("const char* ptr (non-null)",, s, .cstr);
	TEST_AND_PRINT_SET("void * (null)",		, s, .nptr);
	TEST_AND_PRINT_SET("const void * (null)",	, s, .cnptr);
	TEST_AND_PRINT_SET("void * (non-null)",		, s, .ptr);
	TEST_AND_PRINT_SET("const void * (non-null)",	, s, .cptr);


	TEST_AND_PRINT_SET("char",		 , c,   );
	TEST_AND_PRINT_SET("char*",		&, c,   );
	TEST_AND_PRINT_SET("unsigned char",	 , uc,  );
	TEST_AND_PRINT_SET("unsigned char*", 	&, uc,  );
	TEST_AND_PRINT_SET("short",		 , S,   );
	TEST_AND_PRINT_SET("short*",		&, S,   );
	TEST_AND_PRINT_SET("short unsigned",	 , us,  );
	TEST_AND_PRINT_SET("short unsigned*",	&, us,  );
	TEST_AND_PRINT_SET("int",		 , i,   );
	TEST_AND_PRINT_SET("int*",		&, i,   );
	TEST_AND_PRINT_SET("int unsigned",	 , ui,  );
	TEST_AND_PRINT_SET("int unsigned*",	&, ui,  );
	TEST_AND_PRINT_SET("long",		 , l,   );
	TEST_AND_PRINT_SET("long*",		&, l,   );
	TEST_AND_PRINT_SET("long unsigned",	 , ul,  );
	TEST_AND_PRINT_SET("long unsigned*",	&, ul,  );
	TEST_AND_PRINT_SET("long long",		 , ll,  );
	TEST_AND_PRINT_SET("long long*",	&, ll,  );
	TEST_AND_PRINT_SET("long long unsigned", , ull, );
	TEST_AND_PRINT_SET("long long unsigned*",&,ull, );
	TEST_AND_PRINT_SET("float",		 , f,   );
	TEST_AND_PRINT_SET("float*",		&, f,   );
	TEST_AND_PRINT_SET("double",		 , d,   );
	TEST_AND_PRINT_SET("double*",		&, d,   );

	TEST_AND_PRINT_SET("char* defre",	 	, str[0], );
	TEST_AND_PRINT_SET("char* ptr (null)",		, nstr,   );
	TEST_AND_PRINT_SET("char* ptr (non-null)",	, str,    );
	TEST_AND_PRINT_SET("void * (null)",		, nptr,   );
	TEST_AND_PRINT_SET("void * (non-null)",		, ptr,    );

	return 0;
}
